import pytest
import logging

from unittest.mock import AsyncMock

from faststream.rabbit import RabbitBroker, TestRabbitBroker, RabbitQueue

from bovine.activitystreams import factories_for_actor_object
from sqlalchemy import func, select

from cattle_grid.activity_pub.actor import actor_to_object
from cattle_grid.testing.fixtures import *  # noqa
from cattle_grid.database.activity_pub_actor import Follower, Blocking
from cattle_grid.model import ActivityMessage
from cattle_grid.dependencies.globals import global_container


from .outgoing import (
    create_outgoing_router,
    outgoing_message_distribution,
    outgoing_reject_activity,
    outgoing_block_activity,
    outgoing_undo_request,
)

logger = logging.getLogger(__name__)


@pytest.fixture
async def mock_subscriber():
    return AsyncMock()


@pytest.fixture
async def broker(mock_subscriber):
    br = RabbitBroker("amqp://guest:guest@localhost:5672/")
    br.include_router(create_outgoing_router())

    br.subscriber(
        RabbitQueue("test_queue", routing_key="to_send"),
        exchange=global_container.internal_exchange,
    )(mock_subscriber)

    async with TestRabbitBroker(br, connect_only=False) as tbr:
        yield tbr


async def test_outgoing_message_no_call(broker, mock_subscriber):
    try:
        await broker.publish(
            {},
            routing_key="outgoing.Activity",
            exchange=global_container.internal_exchange,
        )
    except Exception as e:
        logger.exception(e)

    mock_subscriber.assert_not_called()


async def test_outgoing_message(broker, mock_subscriber, actor_for_test):
    await broker.publish(
        {
            "actor": actor_for_test.actor_id,
            "data": {"to": "http://remote"},
        },
        routing_key="outgoing.Activity",
        exchange=global_container.internal_exchange,
    )

    mock_subscriber.assert_awaited_once()


async def test_outgoing_message_follower(
    broker, sql_session_for_test, mock_subscriber, actor_for_test
):
    follower_id = "http://follower.test"
    sql_session_for_test.add(
        Follower(
            actor=actor_for_test, follower=follower_id, request="xxx", accepted=True
        )
    )
    await sql_session_for_test.commit()

    await broker.publish(
        {
            "actor": actor_for_test.actor_id,
            "data": {"to": actor_for_test.followers_uri},
        },
        routing_key="outgoing.Activity",
        exchange=global_container.internal_exchange,
    )

    mock_subscriber.assert_awaited_once()
    args = mock_subscriber.call_args

    assert args[1].get("target") == follower_id


async def test_outgoing_follow_request(
    broker, sql_session_for_test, mock_subscriber, actor_for_test
):
    remote_actor = "http://remote"

    activity_factory, _ = factories_for_actor_object(actor_to_object(actor_for_test))

    activity = activity_factory.follow(remote_actor, id="follow_id")

    await broker.publish(
        {"actor": actor_for_test.actor_id, "data": activity},
        routing_key="outgoing.Follow",
        exchange=global_container.internal_exchange,
    )

    await sql_session_for_test.refresh(actor_for_test, attribute_names=["following"])

    assert 1 == len(actor_for_test.following)

    assert not actor_for_test.following[0].accepted
    assert actor_for_test.following[0].following == remote_actor


async def test_outgoing_accept_request(
    broker, sql_session_for_test, mock_subscriber, actor_for_test
):
    follow_request_id = "http://remote/id"

    activity_factory, _ = factories_for_actor_object(actor_to_object(actor_for_test))

    activity = activity_factory.accept(follow_request_id)

    sql_session_for_test.add(
        Follower(
            actor=actor_for_test,
            follower="http://remote",
            request=follow_request_id,
            accepted=False,
        )
    )
    await sql_session_for_test.commit()

    await broker.publish(
        {
            "actor": actor_for_test.actor_id,
            "data": activity,
        },
        routing_key="outgoing.Accept",
        exchange=global_container.internal_exchange,
    )
    await sql_session_for_test.refresh(actor_for_test, attribute_names=["followers"])

    assert 1 == len(actor_for_test.followers)

    assert actor_for_test.followers[0].accepted


async def test_outgoing_no_message_for_public(sql_session_for_test, actor_for_test):
    activity_factory, _ = factories_for_actor_object(actor_to_object(actor_for_test))

    broker = AsyncMock()

    activity = activity_factory.accept("http://remote").as_public().build()

    await outgoing_message_distribution(
        ActivityMessage(
            actor=actor_for_test.actor_id,
            data=activity,
        ),
        session=sql_session_for_test,
        broker=broker,
    )

    broker.publish.assert_not_awaited()


async def test_outgoing_reject(sql_session_for_test, actor_for_test):
    activity_factory, _ = factories_for_actor_object(actor_to_object(actor_for_test))

    broker = AsyncMock()

    sql_session_for_test.add(
        Follower(
            actor=actor_for_test,
            follower="http://remote.test/",
            accepted=True,
            request="http://remote.test/follow",
        )
    )

    await sql_session_for_test.commit()

    activity = activity_factory.reject(
        "http://remote.test/follow", to={"http://remote.test/"}
    ).build()

    await outgoing_reject_activity(
        ActivityMessage(
            actor=actor_for_test.actor_id,
            data=activity,
        ),
        actor=actor_for_test,
        session=sql_session_for_test,
        broker=broker,
    )

    follower_count = await sql_session_for_test.scalar(func.count(Follower.id))

    assert follower_count == 0


async def test_outgoing_block(sql_session_for_test, actor_for_test):
    activity_factory, _ = factories_for_actor_object(actor_to_object(actor_for_test))

    broker = AsyncMock()

    sql_session_for_test.add(
        Follower(
            actor=actor_for_test,
            follower="http://remote.test/",
            accepted=True,
            request="http://remote.test/follow",
        )
    )
    await sql_session_for_test.commit()

    activity = activity_factory.block("http://remote.test/").build()

    await outgoing_block_activity(
        ActivityMessage(
            actor=actor_for_test.actor_id,
            data=activity,
        ),
        session=sql_session_for_test,
        actor=actor_for_test,
        broker=broker,
    )
    await sql_session_for_test.commit()

    follower_count = await sql_session_for_test.scalar(func.count(Follower.id))

    assert follower_count == 0

    blocking_count = await sql_session_for_test.scalar(func.count(Blocking.id))
    assert blocking_count == 1


async def test_outgoing_block_then_undo(sql_session_for_test, actor_for_test):
    activity_factory, _ = factories_for_actor_object(actor_to_object(actor_for_test))

    broker = AsyncMock()
    block_id = "http://me.test/block_id"

    activity = activity_factory.block("http://remote.test/", id=block_id).build()

    await outgoing_block_activity(
        ActivityMessage(
            actor=actor_for_test.actor_id,
            data=activity,
        ),
        session=sql_session_for_test,
        actor=actor_for_test,
        broker=broker,
    )
    await sql_session_for_test.commit()

    blocking_count = await sql_session_for_test.scalar(
        func.count(select(Blocking.id).where(Blocking.active).scalar_subquery())
    )
    assert blocking_count == 1

    undo = activity_factory.undo(activity).build()

    await outgoing_undo_request(
        ActivityMessage(
            actor=actor_for_test.actor_id,
            data=undo,
        ),
        actor=actor_for_test,
        session=sql_session_for_test,
        broker=broker,
    )
    blocking_count = await sql_session_for_test.scalar(
        func.count(select(Blocking.id).where(Blocking.active).scalar_subquery())
    )
    assert blocking_count == 0



import logging


from fastapi import APIRouter, Depends, HTTPException, Request

from cattle_grid.dependencies.fastapi import (
    CommittingSqlSession,
    MethodInformation,
    SqlSession,
)
from cattle_grid.activity_pub.actor import (
    create_actor,
    actor_to_object,
    DuplicateIdentifierException,
)
from cattle_grid.model.account import InformationResponse, EventType

from cattle_grid.database.account import ActorForAccount
from cattle_grid.account.router.info import create_information_response

from cattle_grid.tools import ServerSentEventFromQueueAndTask

from .responses import CreateActorRequest
from .dependencies import CurrentAccount
from .streaming import get_message_streamer

logger = logging.getLogger(__name__)

account_router = APIRouter(prefix="/account", tags=["account"])


@account_router.get(
    "/stream/{event_type}",
    response_description="EventSource",
    operation_id="stream",
)
async def stream(
    event_type: EventType,
    account: CurrentAccount,
    request: Request,
    stream_messages=Depends(get_message_streamer),
):
    """EventSource corresponding to all messages received
    by the account.

    This method returns an
    [EventSource](https://developer.mozilla.org/en-US/docs/Web/API/EventSource)
    providing server sent events."""
    queue, task = stream_messages(account.name, event_type)

    return ServerSentEventFromQueueAndTask(request, queue, task)


@account_router.post(
    "/create",
    status_code=201,
    operation_id="create_actor",
    responses={409: {"description": "Duplicate identifier"}},
)
async def create_actor_method(
    body: CreateActorRequest, account: CurrentAccount, session: CommittingSqlSession
):
    """Allows one to create a new actor. The allowed values for base_url
    can be retrieved using the info endpoint."""
    try:
        actor = await create_actor(
            session, body.base_url, preferred_username=body.handle
        )
    except DuplicateIdentifierException:
        raise HTTPException(409, "Duplicate identifier")

    name = body.name or "from_api"

    session.add(ActorForAccount(account=account, actor=actor.actor_id, name=name))

    return actor_to_object(actor)


@account_router.get("/info", operation_id="account_info")
async def return_account_information(
    account: CurrentAccount, method_information: MethodInformation, session: SqlSession
) -> InformationResponse:
    """Returns information about the server and the account."""

    if not isinstance(method_information, list):
        logger.warning("Method information is not a list")
        method_information = []

    return await create_information_response(session, account, method_information)



import logging

from sqlalchemy import select

from sqlalchemy.ext.asyncio import AsyncSession
from bovine.activitystreams.utils import recipients_for_object, is_public

from cattle_grid.activity_pub.activity import actor_deletes_themselves
from cattle_grid.database.activity_pub_actor import Actor

from .relationship import followers_for_actor, following_for_actor, is_blocked


logger = logging.getLogger(__name__)


class ActorNotFound(Exception):
    """Raised if an actor is not found"""


async def is_valid_requester(
    session: AsyncSession, requester: str, actor: Actor, obj: dict
):
    """Checks if the requested is allowed to view the object

    FIXME tests in test_actor?
    """

    if await is_blocked(session, actor, requester):
        return False

    if is_public(obj):
        return True

    recipients = recipients_for_object(obj)
    self_delete = actor_deletes_themselves(obj)

    recipients = await update_recipients_for_actor(
        session, actor, recipients, self_delete
    )

    valid_requesters = recipients

    if "actor" in obj:
        valid_requesters = valid_requesters | {obj["actor"]}
    if "attributedTo" in obj:
        valid_requesters = valid_requesters | {obj["attributedTo"]}

    return requester in valid_requesters


async def update_recipients_for_actor(
    session: AsyncSession, actor: Actor, recipients, self_delete=False
):
    """Updates set of recipients by removing the followers and following collections, and replacing
    them with the actual sets.

    The following collecting is only allowed for self delete activities.
    """
    if actor.followers_uri in recipients:
        recipients = recipients - {actor.followers_uri} | set(
            await followers_for_actor(session, actor)
        )

        logger.info("Got recipients %s after handling followers", ", ".join(recipients))

    if actor.following_uri in recipients:
        recipients = recipients - {actor.following_uri}

        if self_delete:
            recipients = recipients | set(await following_for_actor(session, actor))
        else:
            logger.warning(
                "Actor '%s' included following collection in recipients where not allowed",
                actor.actor_id,
            )

    return recipients


async def is_valid_requester_for_obj(session: AsyncSession, requester: str, obj: dict):
    """Checks if the requested is allowed to view the object"""

    actor_id = obj.get("attributedTo")
    if actor_id is None:
        actor_id = obj.get("actor")
    if actor_id is None:
        raise ActorNotFound("Object does not have an actor or attributedTo")

    actor = await session.scalar(select(Actor).where(Actor.actor_id == actor_id))
    if actor is None:
        raise ActorNotFound("Actor not found")

    return await is_valid_requester(session, requester, actor, obj)



import pytest
from .property_value import find_key_in_attachments


@pytest.mark.parametrize(
    "attachments, expected",
    [
        ([], None),
        (["string"], None),
        ([{"name": "string"}], None),
        ([{"name": "key"}], None),
        ([{"type": "PropertyValue", "name": "key"}], 0),
        ([{"type": "PropertyValue", "name": "other"}], None),
    ],
)
def test_find_key_in_attachments(attachments, expected):
    assert find_key_in_attachments(attachments, "key") == expected



from typing import List

from fast_depends import inject

from cattle_grid.extensions import Extension
from cattle_grid.model.lookup import LookupMethod, Lookup


def ordered_lookups(extensions: List[Extension]) -> List[LookupMethod]:
    """Returns a list of LookupMethod ordered by lookup order"""
    sorted_extensions = sorted(
        extensions, key=lambda extension: extension.lookup_order or 0, reverse=True
    )

    return [
        extension.lookup_method
        for extension in sorted_extensions
        if extension.lookup_method
    ]


def build_lookup(extensions: list[Extension]) -> LookupMethod:
    """Builds the lookup method"""
    methods = ordered_lookups(extensions)

    async def lookup_result(lookup: Lookup) -> Lookup:
        for method in methods:
            lookup = await inject(method)(lookup)
            if lookup.result is not None:
                return lookup
        return lookup

    return lookup_result



import pytest
from typing import AsyncIterator

import fakeredis
from redis import asyncio as redis


@pytest.fixture
async def redis_client() -> AsyncIterator[redis.Redis]:
    async with fakeredis.FakeAsyncRedis() as client:
        yield client



from pydantic import BaseModel, Field


class CacheConfiguration(BaseModel):
    connection_url: str = Field(default="redis://localhost:6379")
    """Connection url to the key value instance, e.g. redis"""

    duration: int = Field(default=3600)
    """Duration to cache in second"""



import pytest
from unittest.mock import AsyncMock
from faststream.rabbit import RabbitBroker, TestRabbitBroker, RabbitQueue

from cattle_grid.activity_pub.actor import actor_to_object
from cattle_grid.database.activity_pub_actor import PublicIdentifier

from cattle_grid.dependencies.globals import global_container
from cattle_grid.testing.fixtures import *  # noqa

from cattle_grid.model.exchange import DeleteActorMessage, UpdateActorMessage
from cattle_grid.testing import mocked_config
from cattle_grid.model.exchange_update_actor import (
    UpdateIdentifierAction,
    UpdateActionType,
)
from . import create_router


@pytest.fixture
def mock_store_activity():
    return AsyncMock()


@pytest.fixture
async def test_broker(mock_store_activity):
    broker = RabbitBroker()
    broker.include_router(create_router())

    broker.subscriber(
        RabbitQueue("store_activity", routing_key="store_activity"),
        global_container.internal_exchange,
    )(mock_store_activity)

    async with TestRabbitBroker(broker) as tbr:
        yield tbr


async def test_delete_actor(mock_store_activity, test_broker, actor_for_test):
    await test_broker.publish(
        DeleteActorMessage(actor=actor_for_test.actor_id),
        routing_key="delete_actor",
        exchange=global_container.exchange,
    )

    mock_store_activity.assert_awaited_once()


async def test_update_actor(test_broker, sql_session_for_test, actor_for_test):
    new_name = "Alyssa Newton"
    msg = UpdateActorMessage(actor=actor_for_test.actor_id, profile={"name": new_name})

    await test_broker.publish(
        msg,
        routing_key="update_actor",
        exchange=global_container.exchange,
    )

    await sql_session_for_test.refresh(actor_for_test, attribute_names=["profile"])

    obj = actor_to_object(actor_for_test)

    assert obj.get("name") == new_name


async def test_update_actor_profile_not_overwritten(
    sql_session_for_test, test_broker, actor_for_test
):
    new_name = "Alyssa Newton"
    new_description = "I was originally called Leibnitz. Also I'm a cat."
    msg = UpdateActorMessage(actor=actor_for_test.actor_id, profile={"name": new_name})

    await test_broker.publish(
        msg,
        routing_key="update_actor",
        exchange=global_container.exchange,
    )

    await sql_session_for_test.refresh(actor_for_test)

    obj = actor_to_object(actor_for_test)
    assert obj.get("name") == new_name

    msg = UpdateActorMessage(
        actor=actor_for_test.actor_id, profile={"summary": new_description}
    )

    await test_broker.publish(
        msg,
        routing_key="update_actor",
        exchange=global_container.exchange,
    )

    await sql_session_for_test.refresh(actor_for_test)

    obj = actor_to_object(actor_for_test)
    assert obj.get("name") == new_name


async def test_create_identifier(sql_session_for_test, actor_with_account, test_broker):
    with mocked_config({"frontend": {"base_urls": ["http://localhost"]}}):
        identifier = "acct:new@localhost"
        new_identifier = UpdateIdentifierAction(
            action=UpdateActionType.create_identifier,
            identifier=identifier,
            primary=False,
        )

        msg = UpdateActorMessage(
            actor=actor_with_account.actor_id, actions=[new_identifier]
        )

        await test_broker.publish(
            msg,
            routing_key="update_actor",
            exchange=global_container.exchange,
        )
        await sql_session_for_test.refresh(
            actor_with_account, attribute_names=["identifiers"]
        )

        obj = actor_to_object(actor_with_account)

        assert identifier in obj.get("identifiers", [])


async def test_update_identifier(sql_session_for_test, test_broker, actor_for_test):
    identifier = "acct:one@localhost"
    sql_session_for_test.add_all(
        [
            PublicIdentifier(
                actor=actor_for_test,
                identifier="acct:two@localhost",
                name="through_exchange",
                preference=1,
            ),
            PublicIdentifier(
                actor=actor_for_test,
                identifier=identifier,
                name="through_exchange",
                preference=0,
            ),
        ]
    )
    await sql_session_for_test.commit()
    await sql_session_for_test.refresh(actor_for_test, attribute_names=["identifiers"])

    obj = actor_to_object(actor_for_test)

    assert obj["preferredUsername"] == "two"

    update_identifier = UpdateIdentifierAction(
        action=UpdateActionType.update_identifier,
        identifier=identifier,
        primary=True,
    )

    msg = UpdateActorMessage(actor=actor_for_test.actor_id, actions=[update_identifier])

    await test_broker.publish(
        msg,
        routing_key="update_actor",
        exchange=global_container.exchange,
    )

    import asyncio

    await asyncio.sleep(0.0001)

    await sql_session_for_test.refresh(actor_for_test, attribute_names=["identifiers"])

    obj = actor_to_object(actor_for_test)

    assert obj["preferredUsername"] == "one"
